# Tests of SIR under stochastic dynamics
#
# Copyright (C) 2014-2017 Simon Dobson
# 
# This file is part of Complex networks, complex processes (CNCP).
#
# CNCP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CNCP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CNCP. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from .sir import SIRTests

import epyc
import unittest
import networkx

class SIRConsistency(SIRStochasticDynamics):
    '''An extension of the SIRStochasticDynamics class that performs sanity
    checks on the internal data structures after every event.'''
    
    def sanity( self ):
        g = self.network()
        
        # check that all nodes on the infected list are infected
        infecteds = [ n for n in self._infected if g.node[n][self.DYNAMICAL_STATE] != self.INFECTED ]
        if len(infecteds) > 0:
            raise Exception('Un-infected node in infected list')

        # check that all infected nodes are on the infected list
        infecteds = [ n for n in g.nodes_iter() if g.node[n][self.DYNAMICAL_STATE] == self.INFECTED ]
        if set(infecteds) != set(self._infected):
            raise Exception('Infected node not in infected list')

        # check that all SI edges do indeed connect an I to an S (specifically in that order)]
        for (n, m, data) in self._si:
            if (g.node[n][self.DYNAMICAL_STATE] != self.INFECTED) or (g.node[m][self.DYNAMICAL_STATE] != self.SUSCEPTIBLE):
                raise Exception('SI edge not properly formed')

        # check that all SI edges are in the SI edge list
        for n in self._infected:
            for (_, m) in g.edges_iter(n):
                if (g.node[m][self.DYNAMICAL_STATE] == self.SUSCEPTIBLE):
                    if len([ (np, mp, datap) for (np, mp, datap) in self._si if n == np and m == mp ]) == 0:
                        raise Exception('Umanaged SI edge')

    def infect( self, t, params ):
        '''Check sanity of infect event.'''
        ni = len(self._infected)
        super(SIRConsistency, self).infect(t, params)
        if len(self._infected) != ni + 1:
            raise Exception("Infection event didn't generate one new infected node")
        self.sanity()

    def recover( self, t, params ):
        '''Check sanity of recover event.'''
        ni = len(self._infected)
        super(SIRConsistency, self).recover(t, params)
        if len(self._infected) != ni - 1:
            raise Exception("Recovery event didn't recover one infected node")
        self.sanity()

    def at_equilibrium( self, t ):
        '''Check sanity of equilibrium check.'''
        rc = super(SIRConsistency, self).at_equilibrium(t)
        if rc:
            if len(self._si) > 0:
                raise Exception('Termination still leaves SI edges')
        return rc

    
class SIRStochasticTests(SIRTests):

    def setUp( self ):
        super(SIRStochasticTests, self).setUp()
        self._experiment = SIRStochasticDynamics(self._er)
 
    def testConsistency( self ):
        '''Test the data structure consistency over a simulation (slow).'''
        self._experiment.set(self._params).run()
        res = self._experiment.results()



# Gillespie dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import * 
import epyc
import math
import numpy
import networkx


class StochasticDynamics(Dynamics):
    '''A dynamics that runs stochastically in continuous time. This is a
    very efficient and statistically exact approach, but requires that the
    statistical properties of the events making up the process are known.'''
        
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given network.
        
        :param g: prototype network to run the dynamics over (optional)'''
        super(StochasticDynamics, self).__init__(g)

    def transitions( self, t, params ):
        '''Return the transition vector, a sequence of (r, f) pairs
        where r is the rate at which a transition happens and
        f is the transition function called to make it happen. This method
        must be overridden in sub-classes. Note that it's a rate we want,
        not a probability. The former can be obtained from the latter
        simply by multiplying the event probability by the number of
        times it's possible in the network, which is often a population of
        nodes in a given state.
        
        It is perfectly fine for an event to have a zero rate. The process
        is assumed to have reached equilibrium if all events have zero rates.

        It's important that the transitions always come in the same order
        in the vector, even though the rates (and indeed functions) can
        change over time.

        :param t: timestep for which we want the transitions
        :param params: the parameters of the simulation
        :returns: the transition vector'''
        raise NotYetImplementedError('transitions()')
        
    def do( self, params ):
        '''Stochastic dynamics. We use the transition probabilities to
        locate the next event in time, and then call the appropriate
        event service routine.
        
        :param params: parameters of the simulation
        :returns: a dict of simulation properties'''
        rc = dict()

        # run the dynamics
        t = 0
        events = 0
        while not self.at_equilibrium(t):
            # pull the transition dynamics at this timestep
            transitions = self.transitions(t, params)
            
            # compute the total rate of transitions for the entire network
            a = 0.0
            for (r, _) in transitions:
                a = a + r
            if a == 0:       
                break              # no events with non-zero rates 
            
            # calculate the timestep delta
            r1 = numpy.random.random()
            dt = (1.0 / a) * math.log(1.0 / r1)
            t = t + dt
            
            # calculate which transition happens
            if len(transitions) == 1:
                # if there's only one, that's the one that happens
                (_, f) = transitions[0]
            else:
                # otherwise, choose one at random based on the rates
                r2 = numpy.random.random()
                xc = r2 * a
                k = 0
                (xs, f) = transitions[k]
                while xs < xc:
                    k = k + 1
                    (xsp, f) = transitions[k]
                    xs = xs + xsp
            
            # perform the transition
            # call the transition function associated with the selected event
            f(t)
            
            # increment the event counter
            events = events + 1
        
        # record statistics
        rc['timesteps'] = t
        rc['events'] = events
        rc['node_types'] = self.populations()        
        return rc




# Networks dynamics simulation base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epyc
import networkx


class Dynamics(epyc.Experiment, object):
    '''A dynamical process over a network. This is the abstract base class
    for implementing different kinds of dynamics as computational experiments
    suitable for running under. Sub-classes provide synchronous and stochastic
    (Gillespie) simulation dynamics.'''

    # keys for node and edge attributes
    DYNAMICAL_STATE = 'state'   #: dynamical state of a node
    OCCUPIED = 'occupied'       #: edge has been used to transfer infection or not

    # the default maximum simulation time
    DEFAULT_MAX_TIME = 20000
    
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given network.
        The network (if provided) is treated as a prototype that is copied before
        each individual simulation experiment.
        
        :param g: prototype network (optional)'''
        super(Dynamics, self).__init__()
        self._graphPrototype = g
        self._graph = None
        self._maxTime = self.DEFAULT_MAX_TIME

    def network( self ):
        '''Return the network this dynamics is running over.

        :returns: the network'''
        return self._graph

    def setNetworkPrototype( self, g ):
        '''Set the network the dynamics will run over. This will be
        copied for each run of an individual experiment.

        :param g: the network'''
        self._graphPrototype = g

    def setMaximumTime( self, t ):
        '''Set the maximum default simulation time. The default is given
        by `DEFAULT_MAX_TIME`.

        param: t: the maximum time'''
        self._maxTime = t
        
    def at_equilibrium( self, t ):
        '''Test whether the model is an equilibrium. The default runs for
        20000 timesteps and then stops. Override this method to provide
        alternative and/or faster simulations.
        
        :param t: the current simulation timestep
        :returns: True if we're done'''
        return (t >= self._maxTime)

    def setUp( self, params ): 
        '''Before each experiment, create a working copy of the prototype network.

        :param params: parameters of the experiment'''
        self._graph = self._graphPrototype.copy()

    def tearDown( self ):
        '''At the end of each experiment, throw away the copy.'''
        self._graph = None
        
    def skeletonise( self ):
        '''Remove unoccupied edges from the network. This leaves the network
        consisting of only "occupied" edges that were used to transmit the
        infection between nodes.
        
        :returns: the network with unoccupied edges removed'''
        
        # find all unoccupied edges
        g = self.network()
        edges = []
        for n in g.nodes_iter():
            for (np, m, data) in g.edges_iter(n, data = True):
                if (self.OCCUPIED not in data.keys()) or (data[self.OCCUPIED] != True):
                    # edge is unoccupied, mark it to be removed
                    # (safe because there are no parallel edges)
                    edges.insert(0, (n, m))
                    
        # remove all the unoccupied edges in one go
        g.remove_edges_from(edges)
        return g
    
    def populations( self ):
        '''Return a count of the number of nodes in each dynamical state.
        
        :returns: a dict mapping states to number of nodes in that state'''
        g = self.network()
        pops = dict()
        for n in g.nodes_iter():
            s = g.node[n][self.DYNAMICAL_STATE]
            if s not in pops.keys():
                pops[s] = 1
            else:
                pops[s] = pops[s] + 1
        return pops
    



# SIR simulator with stochastic (Gillespie) dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import epyc
import networkx
import numpy


class SIRStochasticDynamics(StochasticDynamics):
    '''A stochastic SIR dynamics.'''

    # the possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE = 'S'    #: Node is susceptible to infection
    INFECTED = 'I'       #: Node is infected
    RECOVERED = 'R'      #: Node is recovered/removed
    
    def __init__( self, g = None ):
        '''Generate an SIR dynamics.
        
        :param g: the graph to run over (optional)'''
        super(SIRStochasticDynamics, self).__init__(g)

        # list of infected nodes, the sites of all the dynamics
        self._infected = []
    
        # list of SI edges connecting a susceptible to an infected node
        self._si = []

    def setUp( self, params ):
        '''Seed the network with infected nodes, extract the initial set of
        SI nodes, and mark all edges as unoccupied by the dynamics.

        :param params: the paemeters of the experiment'''

        # perform the default setup
        super(SIRStochasticDynamics, self).setUp(params)

        g = self.network()
        pInfected = params['pInfected']

        # in case we re-run from a dirty intermediate state
        self._infected = []       
        self._si = []
        
        # infect nodes
        for n in g.node.keys():
            if numpy.random.random() <= pInfected:
                self._infected.insert(0, n)
                g.node[n][self.DYNAMICAL_STATE] = self.INFECTED
            else:
                g.node[n][self.DYNAMICAL_STATE] = self.SUSCEPTIBLE
                
        # extract the initial set of SI edges
        for (n, m, data) in g.edges_iter(self._infected, data = True):
            if g.node[m][self.DYNAMICAL_STATE] == self.SUSCEPTIBLE:
                self._si.insert(0, (n, m, data))
        
        # mark all edges as unoccupied
        for (n, m, data) in g.edges_iter(data = True):
            data[self.OCCUPIED] = False
            
    def at_equilibrium( self, t ):
        '''SIR dynamics is at equilibrium if there are no more infected nodes left
        in the network, no susceptible nodes adjacent to infected nodes, or if we've
        exceeded the default simulation length.
        
        :param t: the current time
        :returns: True if the model has stopped'''
        if (len(self._infected) == 0):
            return True
        else:
            return super(SIRStochasticDynamics, self).at_equilibrium(t)
         
    def infect( self, t, params ):
        '''Infect a node chosen at random from the SI edges.

        :param t: the timestep
        :param params: the parameters of the experiment'''
        g = self.network()
        
        # choose an SI edge
        i = int(numpy.random.random() * (len(self._si) - 1))
        (n, m, data) = self._si[i]
        
        # infect the susceptible end
        self._infected.insert(0, m)
        g.node[m][self.DYNAMICAL_STATE] = self.INFECTED
        
        # label the edge we traversed as occupied
        data[self.OCCUPIED] = True
        
        # remove all edges in the SI list from an infected node to the
        # node we just infected
        self._si = [ (np, mp, data) for (np, mp, data) in self._si if mp != m ]
        
        # add all the edges incident on this node connected to susceptible nodes
        for (_, mp, datap) in g.edges_iter(m, data = True):
            if g.node[mp][self.DYNAMICAL_STATE] == self.SUSCEPTIBLE:
                self._si.insert(0, (m, mp, datap))

    def recover( self, t, params ):
        '''Cause a node to recover.

        :param t: the timestep
        :param params: the parameters of the experiment'''
        g = self.network()
        
        # choose an infected node at random
        i = int(numpy.random.random() * (len(self._infected) - 1))
        n = self._infected[i]
        
        # mark the node as recovered
        del self._infected[i]
        g.node[n][self.DYNAMICAL_STATE] = self.RECOVERED
        
        # remove all edges in the SI list incident on this node
        self._si = [ (np, m, datap) for (np, m, datap) in self._si if np != n ]
        
    def transitions( self, t, params ):
        '''Return the transition vector for the dynamics.
        
        :param t: time (ignored)
        :param params: the parameters of the simulation
        :returns: the transition vector'''
        
        # transitions are expressed as rates, whereas we're specified
        # in terms of probabilities, so we convert the latter to the former.
        return [ (len(self._si) * params['pInfect'],        lambda t: self.infect(t, params)),
                 (len(self._infected) * params['pRecover'], lambda t: self.recover(t, params)) ]
            
    def do( self, params ):
        '''Returns statistics of outbreak sizes. This skeletonises the
        network, so it can't have any further dynamics run on it.
        
        :param params: the parameters of the experiment
        :returns: a dict of statistical properties'''
        g = self.network()
        
        # run the basic dynamics
        rc = super(SIRStochasticDynamics, self).do(params)
        
        # compute the limits and means
        cs = sorted(networkx.connected_components(self.skeletonise()), key = len, reverse = True)
        max_outbreak_size = len(cs[0])
        max_outbreak_proportion = (max_outbreak_size + 0.0) / g.order()
        mean_outbreak_size = numpy.mean([ len(c) for c in cs ])
        
        # add metrics for this simulation run
        rc['mean_outbreak_size'] = mean_outbreak_size,
        rc['max_outbreak_size'] = max_outbreak_size,
        rc['max_outbreak_proportion'] = max_outbreak_proportion
        return rc



# Synchronous dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import epyc
import networkx


class SynchronousDynamics(Dynamics):
    '''A dynamics that runs synchronously in discrete time, applying local
    rules to each node in the network. These are simple to understand and
    simple to code for many cases, but can be statistically inexact and slow
    for large systems.'''
        
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given prototype
        network.
        
        :param g: prototype network to run over (optional)'''
        super(SynchronousDynamics, self).__init__(g)
        
    def model( self, n, params ):
        '''The dynamics function that's run over the network. This
        is a placeholder to be re-defined by sub-classes.
        
        :param params: the parameters of the simulation
        :param n: the node being simulated'''
        raise NotYetImplementedError('model()')

    def dynamics( self, t, params ):
        '''Run a single step of the model over the network.
        
        :param t: the current timestep
        :param params: the parameters of the simulation
        :returns: the number of dynamic events that happened in this timestep'''
        g = self.network()
        events = 0
        for n in g.node.keys():
            events = events + self.model(n, params)
        return events

    def do( self, params ):
        '''Synchronous dynamics. We apply dynamics() at each timestep
        and then check for completion using at_equilibrium().
        
        :param params: the parameters of the simulation
        :returns: a dict of simulation properties'''
        rc = dict()

        t = 0
        events = 0
        timestepEvents = 0
        while not self.at_equilibrium(t):
            t = t + 1

            # run a step
            nev = self.dynamics(t, params)
            if nev > 0:
                events = events + nev
                timestepEvents = timestepEvents + 1
        
        # return the simulation-level results
        rc['timesteps'] = t
        rc['events'] = events
        rc['timesteps_with_events'] = timestepEvents
        rc['node_types'] = self.populations()
        return rc




# SIS simulator with synchronous dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.#

from epydemic import *
import epyc
import networkx
import numpy


class SISSynchronousDynamics(SIRSynchronousDynamics):
    '''An SIR synchronous dynamics.'''
    
    def __init__( self, g = None ):
        '''Generate an SIS dynamics over an (optional) prototype network.
        
        :param g: the graph to run over (optional)'''
        super(SISSynchronousDynamics, self).__init__(g)

        # list of infected nodes, the sites of all the dynamics
        self._infected = []

    def model( self, n, params ):
        '''Apply the SIS dynamics to node n. Nodes "recover" back to the
        susceptible state.

        :param params: the parameters of thje simulation
        :param n: the node
        :returns: the number of changes made'''
        g = self.network()
        pInfect = params['pInfect']
        pRecover = params['pRecover']
        events = 0
        
        # infect susceptible neighbours with probability pInfect
        for (_, m, data) in g.edges_iter(n, data = True):
            if g.node[m][self.DYNAMICAL_STATE] == self.SUSCEPTIBLE:
                # we've got a susceptible neighbour, do we infect them?
                if numpy.random.random() <= pInfect:
                    # yes we do!
                    events = events + 1
                    
                    # infect the node
                    g.node[m][self.DYNAMICAL_STATE] = self.INFECTED
                    self._infected.insert(0, m)
                        
                    # label the edge we traversed as occupied
                    data[self.OCCUPIED] = True
    
        # recover with probability pRecover
        if numpy.random.random() <= pRecover:
            # recover the node
            events = events + 1
            g.node[n][self.DYNAMICAL_STATE] = self.SUSCEPTIBLE   # note that this potentially allows 
                                                                 # re-infection in this timestep
     
        return events



# SIS simulator with stochastic (Gillespie) dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.#

from epydemic import *
import epyc
import networkx
import numpy


class SISStochasticDynamics(SIRStochasticDynamics):
    '''An SIS stochastic dynamics. This is very like the SIR
    dynamics but recovered nodes become susceptible again, modelling
    recurrent disease where infection does not confer immunity. SIS also
    allows endemic diseases to form, existing indefinitely in the network
    at low levels.'''

    def __init__( self, g = None ):
        '''Generate an SIS dynamics with an optional prototype network.
        
        :param g: the graph to run over (optional)'''
        super(SISStochasticDynamics, self).__init__(g)

    def recover( self, t, params ):
        '''Cause a node to recover back to the susceptible state.

        :param t: the timestep
        :param params: the parameters of the experiment'''
        g = self.network()
        
        # choose an infected node at random
        i = int(numpy.random.random() * (len(self._infected) - 1))
        n = self._infected[i]
        
        # mark the node as recovered
        del self._infected[i]
        g.node[n][self.DYNAMICAL_STATE] = self.SUSCEPTIBLE
        
        # remove all edges in the SI list incident on this node
        self._si = [ (np, m, e) for (np, m, e) in self._si if np != n ]
 



# Tests of SIR under synchronous dynamics
#
# Copyright (C) 2014-2017 Simon Dobson
# 
# This file is part of Complex networks, complex processes (CNCP).
#
# CNCP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CNCP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CNCP. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from .sir import SIRTests

import epyc
import unittest
import networkx


class SIRSynchronousTests(SIRTests):

    def setUp( self ):
        super(SIRSynchronousTests, self).setUp()
        self._experiment = SIRSynchronousDynamics(self._er)

 



# SIR simulator with synchronous dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
import epyc
import networkx
import numpy
from copy import copy


class SIRSynchronousDynamics(SynchronousDynamics):
    '''A synchronous SIR dynamics. We use probabilities
    to express infection and recovery per timestep, and run the system
    using synchronous dynamics.'''
    
    # the possible dynamics states of a node for SIR dynamics
    SUSCEPTIBLE = 'S'    #: Node is susceptible to infection
    INFECTED = 'I'       #: Node is infected
    RECOVERED = 'R'      #: Node is recovered/removed
        
    def __init__( self, g = None ):
        '''Generate an SIR dynamics over an (optional) prototype network.
        
        :param g: the graph to run over (optional)'''
        super(SIRSynchronousDynamics, self).__init__(g)

        # list of infected nodes, the sites of all the dynamics
        self._infected = []

    def setUp( self, params ):
        '''Seed the network with infected nodes, and mark all edges
        as unoccupied by the dynamics.

        :param params: the parameters of the simulation'''

        # perform the base behaviour, which copies the prototype network
        super(SIRSynchronousDynamics, self).setUp(params)
        
        g = self.network()
        pInfected = params['pInfected']
        
        # in case we re-run from a dirty intermediate state
        self._infected = []
        
        # seed the network with infected nodes, making all the other susceptible
        for n in g.node.keys():
            if numpy.random.random() <= pInfected:
                self._infected.insert(0, n)
                g.node[n][self.DYNAMICAL_STATE] = self.INFECTED
            else:
                g.node[n][self.DYNAMICAL_STATE] = self.SUSCEPTIBLE
                
        # mark all edges as unoccupied
        for (n, m, data) in g.edges_iter(data = True):
            data[self.OCCUPIED] = False

    def dynamics( self, t, params ):
        '''Optimised per-step dynamics that only runs the model at infected
        nodes, since they're the only places where state changes originate. At the
        end of each timestep we re-build the infected node list.
        
        :param t: the timestep
        :param params: the parameters of the simulation
        :returns: the number of events that happened in this timestep'''
        g = self.network()
        events = 0
        
        # run model dynamics on infected nodes only
        for n in copy(self._infected):
            events = events + self.model(n, params)
    
        # retain only infected nodes in the infected list
        self._infected = [ n for n in self._infected if g.node[n][self.DYNAMICAL_STATE] == self.INFECTED ]
        
        return events
            
    def model( self, n, params ):
        '''Apply the SIR dynamics to node n. From the re-definition of
        :meth:`SIRSynchronousDynamics.dynamics` we already know this node is infected.

        :param params: the parameters of the simulation
        :param n: the node
        :returns: the number of changes made'''
        g = self.network()
        pInfect = params['pInfect']
        pRecover = params['pRecover']
        events = 0
        
        # infect susceptible neighbours with probability pInfect
        for (_, m, data) in g.edges_iter(n, data = True):
            if g.node[m][self.DYNAMICAL_STATE] == self.SUSCEPTIBLE:
                # we've got a susceptible neighbour, do we infect them?
                if numpy.random.random() <= pInfect:
                    # yes we do!
                    events = events + 1
                    
                    # infect the node
                    g.node[m][self.DYNAMICAL_STATE] = self.INFECTED
                    self._infected.insert(0, m)
                        
                    # label the edge we traversed as occupied
                    data[self.OCCUPIED] = True
    
        # recover with probability pRecover
        if numpy.random.random() <= pRecover:
            # recover the node
            events = events + 1
            g.node[n][self.DYNAMICAL_STATE] = self.RECOVERED
                
        return events
            
    def at_equilibrium( self, t ):
        '''SIR dynamics is at equilibrium if there are no more infected nodes
        left in the network or if we've hit the default equilibrium conditions.
        
        :returns: True if the model has stopped'''
        if (len(self._infected) == 0):
            return True
        else:
            return super(SIRSynchronousDynamics, self).at_equilibrium(t)
            
    def do( self, params ):
        '''Run the dynamics and return statistics of outbreak sizes.
        
        :param params: parameters of the simulation
        :returns: a dict of statistical properties'''
        
        # run the basic dynamics
        rc = super(SIRSynchronousDynamics, self).do(params)
        
        # compute the limits and means
        cs = sorted(networkx.connected_components(self.skeletonise()), key = len, reverse = True)
        max_outbreak_size = len(cs[0])
        max_outbreak_proportion = (max_outbreak_size + 0.0) / self.network().order()
        mean_outbreak_size = numpy.mean([ len(c) for c in cs ])
        
        # add  metrics for this simulation run
        rc['mean_outbreak_size'] = mean_outbreak_size,
        rc['max_outbreak_size'] = max_outbreak_size,
        rc['max_outbreak_proportion'] = max_outbreak_proportion
        return rc
    



# Base class for tests of SIR under various simulation dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import unittest
import networkx


class SIRTests(unittest.TestCase):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        
        # single experiment
        self._params = dict(pInfect = 0.01,
                            pInfected = 0.001,
                            pRecover = 0.02)
        self._er = networkx.erdos_renyi_graph(5000, 0.004)

        # lab run
        self._lab = epyc.Lab()
        self._lab['pInfect'] = [ 0.01 ]
        self._lab['pInfected'] = [ 0.001, 0.002, 0.005 ]
        self._lab['pRecover'] = [ 0.02 ]
 
    
    def testSingle( self ):
        '''Test a single run of SIR.'''
        self._experiment.set(self._params).run()
        res = self._experiment.results()
        #print res
        
        self.assertIn('timesteps', res[epyc.Experiment.RESULTS])
        self.assertIn('events', res[epyc.Experiment.RESULTS])
        self.assertTrue(res[epyc.Experiment.RESULTS]['events'] > 0)
        self.assertIn('mean_outbreak_size', res[epyc.Experiment.RESULTS])

    def testLab( self ):
        '''Test successful SIR experiments across a parameter space.'''       
        self._lab.runExperiment(self._experiment)
        rcs = self._lab.results()
        #print rcs

        self.assertTrue(len(rcs), 3)
        for res in rcs:
            self.assertIn('timesteps', res[epyc.Experiment.RESULTS])
            self.assertIn('events', res[epyc.Experiment.RESULTS])
            self.assertTrue(res[epyc.Experiment.RESULTS]['events'] > 0)
            self.assertIn('mean_outbreak_size', res[epyc.Experiment.RESULTS])

    def testInfection( self ):
        '''Test that an experiment converges.'''
        self._experiment.set(self._params).run()
        res = self._experiment.results()
        #print res
        
        self.assertTrue(res[epyc.Experiment.RESULTS]['mean_outbreak_size'] > 0.5)
        
        
